/*========== Includes ==========*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <dirent.h>
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#include <windows.h>
#undef PlaySound
#endif

#include <SDL.h>
#include <SDL_audio.h>
#include <SDL_thread.h>
#include <SDL_ttf.h>
#include <SDL_image.h>
/*========== Includes ==========*/

/*========== Defines ==========*/
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#define SLASH "\\"
#else
#define SLASH "/"
#endif

#define PNG_VARIOUS "png" SLASH "various" SLASH
#define PNG_BUTTONS "png" SLASH "buttons" SLASH
#define PNG_GAMEPAD "png" SLASH "gamepad" SLASH
#define PNG_LIVING "png" SLASH "living" SLASH
#define PNG_SLIVING "png" SLASH "sliving" SLASH
#define PNG_EXTRAS "png" SLASH "extras" SLASH
#define PNG_L0003 "png" SLASH "l0003" SLASH

#define EXIT_NORMAL 0
#define EXIT_ERROR 1
#define EDITOR_NAME "TODO"
#define EDITOR_VERSION "TODO"
#define COPYRIGHT "TODO"
#define DISKA "fdi" SLASH "DISK_A.FDI"
#define MAX_FILENAME 20
#define MAX_DATA 720
#define ENTRIES 256
#define MAX_FILEDATA 90000 /*** TODO ***/
#define MAX_OPTION 100
#define MAX_LEVEL 15
#define ROOMS 24
#define TILES 30
#define BLOCKS 128
#define WINDOW_WIDTH 690
#define WINDOW_HEIGHT 467
#define MAX_IMG 200
#define MAX_CON 30
#define NUM_SOUNDS 20 /*** Sounds that may play at the same time. ***/
#define BAR_FULL 443
#define MAX_TEXT 100
#define REFRESH 25 /*** That is 40 frames per second, 1000/25. ***/

#ifndef O_BINARY
#define O_BINARY 0
#endif
/*========== Defines ==========*/

int iFdA;
int iDebug;
int iNoAudio;
int iScale;
int iFullscreen;
int iStartLevel;
int iNoController;
int iBrokenRoomLinks;
int iDone[24 + 2];
int iHeader;
int iNoAnim;
int iCurRoom;
SDL_Window *window;
unsigned int iWindowID;
SDL_Renderer *ascreen;
unsigned int iActiveWindowID;
SDL_Cursor *curArrow;
SDL_Cursor *curWait;
SDL_Cursor *curHand;
int iHourglassFrame;
int iSandFrame;
int iPreLoaded;
int iNrToPreLoad;
int iCurrentBarHeight;
TTF_Font *font11;
TTF_Font *font15;
TTF_Font *font20;
int iStatusBarFrame;
int iScreen;
int iXPos, iYPos;
int iChanged;
int iCurLevel;
int iDownAt;
int iSelected;
int iExtras;
int iOnTile;
int iChangeEvent;
int iPlaytest;

char sFileNames[ENTRIES + 2][MAX_FILENAME + 2];
int iFirstSectors[ENTRIES + 2];

unsigned char sFileCom[MAX_FILEDATA + 2];
unsigned char sFileDecom[MAX_FILEDATA + 2];
int iSizeCom;
int iSizeDecom;
int iPosIn;

/*** level ***/
int iKidRoom, iKidPos, iKidDir;
int arBlock[ROOMS + 2][TILES + 2];
int arBlockFlag[BLOCKS + 2];
int arBlockObject[BLOCKS + 2];
int arRoomLinks[ROOMS + 2][4 + 2];

int arRoomLinksBroken[ROOMS + 2][4 + 2];
int iMinX, iMaxX, iMinY, iMaxY;

unsigned int gamespeed;
Uint32 looptime;

/*** controller ***/
int iController;
SDL_GameController *controller;
char sControllerName[MAX_CON + 2];
SDL_Joystick *joystick;
SDL_Haptic *haptic;

/*** for text ***/
SDL_Surface *message;
SDL_Texture *messaget;
SDL_Rect offset;
SDL_Color color_bl = {0x00, 0x00, 0x00, 255};
SDL_Color color_wh = {0xff, 0xff, 0xff, 255};

SDL_Texture *imgloading;
SDL_Texture *imghourglasssprite;
SDL_Texture *imgsandsprite;
SDL_Texture *imgblack;
SDL_Texture *imgleft_0, *imgleft_1;
SDL_Texture *imgright_0, *imgright_1;
SDL_Texture *imgup_0, *imgup_1;
SDL_Texture *imgdown_0, *imgdown_1;
SDL_Texture *imglrno, *imgudno, *imgudnonfo;
SDL_Texture *imgroomsoff, *imgroomson_0, *imgroomson_1;
SDL_Texture *imgbroomsoff, *imgbroomson_0, *imgbroomson_1;
SDL_Texture *imgeventsoff, *imgeventson_0, *imgeventson_1;
SDL_Texture *imgsaveoff, *imgsaveon_0, *imgsaveon_1;
SDL_Texture *imgquit_0, *imgquit_1;
SDL_Texture *imgprevoff, *imgprevon_0, *imgprevon_1;
SDL_Texture *imgnextoff, *imgnexton_0, *imgnexton_1;
SDL_Texture *imgkidr[2 + 2], *imgkidl[2 + 2];
SDL_Texture *imgbar;
SDL_Texture *extras[10 + 2];
SDL_Texture *imgunk;
SDL_Texture *imgplaytest;
SDL_Texture *imgl0003[255 + 2];
SDL_Texture *imgselblock;

struct sample {
	Uint8 *data;
	Uint32 dpos;
	Uint32 dlen;
} sounds[NUM_SOUNDS];

void ShowUsage (void);
void GetOptionValue (char *sArgv, char *sValue);
void LoadFileCom (int iFd, char *sFileName);
void LSeek (int iFd, int iOffset);
int ReadFromFile (int iFd, char *sWhat, int iSize, unsigned char *sRetString);
void WriteCharByChar (int iFd, unsigned char *sString, int iLength);
void PrIfDe (char *sString);
void LoadFileNamesAndFirstSectors (int iFd);
void DecompressFile (void);
void DecompFun (int iType, int cPos1, int cPos2, int cPos3, int cPos4);
int CheckHeader (int iFd, int iOffset);
unsigned char NibbleLower (unsigned char cChar);
unsigned char NibbleUpper (unsigned char cChar);
unsigned char NibbleMerge (unsigned char cUpper, unsigned char cLower);
void SaveFileDecom (char *sFileName);
void LoadLevel (int iLevel);
void InitScreenAction (char *sAction);
void InitScreen (void);
void Quit (void);
void GetBlockName (int iBlock, char *sName);
void GetFlagName (int iFlag, char *sName);
void GetObjectName (int iObject, char *sName);
char cShowDirection (int iDirection);
void LoadFonts (void);
void MixAudio (void *unused, Uint8 *stream, int iLen);
void PreLoad (char *sPath, char *sPNG, SDL_Texture **imgImage);
void ShowImage (SDL_Texture *img, int iX, int iY, char *sImageInfo,
	SDL_Renderer *screen, float fMultiply, int iXYScale);
void LoadingBar (int iBarHeight);
void ShowScreen (int iScreenS);
void InitPopUp (void);
void InitPopUpSave (void);
void ShowTile (int iBlock, int iLocation);
void LocationToXY (int iLocation, int iXY[4]);
int InArea (int iUpperLeftX, int iUpperLeftY,
	int iLowerRightX, int iLowerRightY);
void PlaySound (char *sFile);
int BrokenRoomLinks (int iPrint);
void CheckSides (int iRoom, int iX, int iY);
void Prev (void);
void Next (void);
void DisplayText (int iStartX, int iStartY, int iFontSize,
	char arText[9 + 2][MAX_TEXT + 2], int iLines, TTF_Font *font,
	SDL_Color back, SDL_Color fore);
void CustomRenderCopy (SDL_Texture* src, char *sImageInfo, SDL_Rect* srcrect,
	SDL_Renderer* dst, SDL_Rect *dstrect);
void ChangeEvent (int iAmount, int iChangePos);
int OnLevelBar (void);

/*****************************************************************************/
int main (int argc, char *argv[])
/*****************************************************************************/
{
	int iLoopArg;
	char sStartLevel[MAX_OPTION + 2];
	time_t tm;
	SDL_version verc, verl;

	iDebug = 0;
	iNoAudio = 0;
	iScale = 1;
	iFullscreen = 0;
	iStartLevel = 1;
	iNoController = 0;
	iNoAnim = 0;

	if (argc > 1)
	{
		for (iLoopArg = 1; iLoopArg <= argc - 1; iLoopArg++)
		{
			if ((strcmp (argv[iLoopArg], "-h") == 0) ||
				(strcmp (argv[iLoopArg], "-?") == 0) ||
				(strcmp (argv[iLoopArg], "--help") == 0))
			{
				ShowUsage();
			}
			else if ((strcmp (argv[iLoopArg], "-v") == 0) ||
				(strcmp (argv[iLoopArg], "--version") == 0))
			{
				printf ("%s %s\n", EDITOR_NAME, EDITOR_VERSION);
				exit (EXIT_NORMAL);
			}
			else if ((strcmp (argv[iLoopArg], "-d") == 0) ||
				(strcmp (argv[iLoopArg], "--debug") == 0))
			{
				iDebug = 1;
			}
			else if ((strcmp (argv[iLoopArg], "-n") == 0) ||
				(strcmp (argv[iLoopArg], "--noaudio") == 0))
			{
				iNoAudio = 1;
			}
			else if ((strcmp (argv[iLoopArg], "-z") == 0) ||
				(strcmp (argv[iLoopArg], "--zoom") == 0))
			{
				iScale = 2;
			}
			else if ((strcmp (argv[iLoopArg], "-f") == 0) ||
				(strcmp (argv[iLoopArg], "--fullscreen") == 0))
			{
				iFullscreen = SDL_WINDOW_FULLSCREEN_DESKTOP;
			}
			else if ((strncmp (argv[iLoopArg], "-l=", 3) == 0) ||
				(strncmp (argv[iLoopArg], "--level=", 8) == 0))
			{
				GetOptionValue (argv[iLoopArg], sStartLevel);
				iStartLevel = atoi (sStartLevel);
				if ((iStartLevel < 0) || (iStartLevel > MAX_LEVEL))
				{
					iStartLevel = 1;
				}
			}
			else if ((strcmp (argv[iLoopArg], "-s") == 0) ||
				(strcmp (argv[iLoopArg], "--static") == 0))
			{
				iNoAnim = 1;
			}
			else if ((strcmp (argv[iLoopArg], "-k") == 0) ||
				(strcmp (argv[iLoopArg], "--keyboard") == 0))
			{
				iNoController = 1;
			}
			else
			{
				ShowUsage();
			}
		}
	}

	iFdA = open (DISKA, O_RDONLY|O_BINARY);
	if (iFdA == -1)
	{
		printf ("[FAILED] Could not open \"%s\": %s!\n",
			DISKA, strerror (errno));
		exit (EXIT_ERROR);
	}

	/*** iHeader ***/
	if (CheckHeader (iFdA, 0x00) == 1)
	{
		iHeader = 0;
	} else if (CheckHeader (iFdA, 0x1000) == 1) {
		iHeader = 1;
	} else {
		printf ("[FAILED] Cannot find header bytes!\n");
		exit (EXIT_ERROR);
	}

	LoadFileNamesAndFirstSectors (iFdA);

	srand ((unsigned)time(&tm));

	LoadLevel (iStartLevel);

	/*** Show the SDL version used for compiling and linking. ***/
	if (iDebug == 1)
	{
		SDL_VERSION (&verc);
		SDL_GetVersion (&verl);
		printf ("[ INFO ] Compiled with SDL %u.%u.%u, linked with SDL %u.%u.%u.\n",
			verc.major, verc.minor, verc.patch, verl.major, verl.minor, verl.patch);
	}

	InitScreen();

	Quit();

	return 0;
}
/*****************************************************************************/
void ShowUsage (void)
/*****************************************************************************/
{
	printf ("%s %s\n%s\n\n", EDITOR_NAME, EDITOR_VERSION, COPYRIGHT);
	printf ("Usage:\n");
	printf ("  %s [OPTIONS]\n\nOptions:\n", EDITOR_NAME);
	printf ("  -h, -?,    --help           display this help and exit\n");
	printf ("  -v,        --version        output version information and"
		" exit\n");
	printf ("  -d,        --debug          also show levels on the console\n");
	printf ("  -n,        --noaudio        do not play sound effects\n");
	printf ("  -z,        --zoom           double the interface size\n");
	printf ("  -f,        --fullscreen     start in fullscreen mode\n");
	printf ("  -l=NR,     --level=NR       start in level NR\n");
	printf ("  -s,        --static         do not display animations\n");
	printf ("  -k,        --keyboard       do not use a game controller\n");
	printf ("\n");
	exit (EXIT_NORMAL);
}
/*****************************************************************************/
void GetOptionValue (char *sArgv, char *sValue)
/*****************************************************************************/
{
	int iTemp;
	char sTemp[MAX_OPTION + 2];
  
	iTemp = strlen (sArgv) - 1;
	snprintf (sValue, MAX_OPTION, "%s", "");
	while (sArgv[iTemp] != '=')
	{
		snprintf (sTemp, MAX_OPTION, "%c%s", sArgv[iTemp], sValue);
		snprintf (sValue, MAX_OPTION, "%s", sTemp);
		iTemp--;
	}
}
/*****************************************************************************/
void LoadFileCom (int iFd, char *sFileName)
/*****************************************************************************/
{
	int iCurrentSector, iNextSector;
	int iExit;
	int iOffset;
	unsigned char sNextSector[2 + 2];
	int iSize;
	unsigned char sSectorData[0x400 + 2];

	/*** Used for looping. ***/
	int iLoopEntry;
	int iLoopByte;

	sFileCom[0] = '\0';
	iSizeCom = 0;

	/*** iCurrentSector ***/
	for (iLoopEntry = 1; iLoopEntry <= ENTRIES; iLoopEntry++)
	{
		if (strcmp (sFileNames[iLoopEntry], sFileName) == 0)
			{ iCurrentSector = iFirstSectors[iLoopEntry]; }
	}
	if (iDebug == 1)
		{ printf ("[ INFO ] Current sector: %i\n", iCurrentSector); }

	iExit = 0;
	do {
		if ((iCurrentSector < 0x0008) || (iCurrentSector > 0x04CF))
		{
			printf ("[ WARN ] Invalid sector: %i!\n", iCurrentSector);
		}

		/*** iNextSector ***/
		iOffset = 0x400 + (iCurrentSector * 2);
		LSeek (iFd, iOffset);
		ReadFromFile (iFd, "Next Sector", 2, sNextSector);
		iNextSector = sNextSector[0] + (sNextSector[1] << 8);
		if (iDebug == 1)
			{ printf ("[ INFO ] Next sector: %i\n", iNextSector); }

		iOffset = iCurrentSector * 0x400;
		LSeek (iFd, iOffset);
		ReadFromFile (iFd, "Sector Data", 0x400, sSectorData);
		if (iNextSector < 0xFC00)
		{
			iSize = 0x400; /*** Everything. ***/
			for (iLoopByte = 0; iLoopByte < iSize; iLoopByte++)
			{
				sFileCom[iSizeCom] = sSectorData[iLoopByte];
				iSizeCom++;
			}
			iCurrentSector = iNextSector;
		} else {
			iSize = iNextSector - 0xFC00;
			if (iSize == 0) { iSize = 0x400; }
			for (iLoopByte = 0; iLoopByte < iSize; iLoopByte++)
			{
				sFileCom[iSizeCom] = sSectorData[iLoopByte];
				iSizeCom++;
			}
			iExit = 1;
		}
	} while (iExit == 0);
	sFileCom[iSizeCom] = '\0';
}
/*****************************************************************************/
void LSeek (int iFd, int iOffset)
/*****************************************************************************/
{
	lseek (iFd, iOffset + (iHeader * 0x1000), SEEK_SET);
}
/*****************************************************************************/
int ReadFromFile (int iFd, char *sWhat, int iSize, unsigned char *sRetString)
/*****************************************************************************/
{
	int iLength;
	int iRead;
	char sRead[1 + 2];
	int iEOF;

	if ((iDebug == 1) && (strcmp (sWhat, "") != 0))
	{
		printf ("[  OK  ] Loading: %s\n", sWhat);
	}
	iLength = 0;
	iEOF = 0;
	do {
		iRead = read (iFd, sRead, 1);
		switch (iRead)
		{
			case -1:
				printf ("[FAILED] Could not read: %s!\n", strerror (errno));
				exit (EXIT_ERROR);
				break;
			case 0: PrIfDe ("[ INFO ] End of file\n"); iEOF = 1; break;
			default:
				sRetString[iLength] = sRead[0];
				iLength++;
				break;
		}
	} while ((iLength < iSize) && (iEOF == 0));
	sRetString[iLength] = '\0';

	return (iLength);
}
/*****************************************************************************/
void WriteCharByChar (int iFd, unsigned char *sString, int iLength)
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iPos;

	for (iPos = 0; iPos < iLength; iPos++)
	{
		write (iFd, (char *)&sString[iPos], 1);
	}
}
/*****************************************************************************/
void PrIfDe (char *sString)
/*****************************************************************************/
{
	if (iDebug == 1) { printf ("%s", sString); }
}
/*****************************************************************************/
void LoadFileNamesAndFirstSectors (int iFd)
/*****************************************************************************/
{
	unsigned char sBaseName[8 + 2];
	unsigned char sExtension[3 + 2];
	unsigned char sUnknown[3 + 2];
	unsigned char sFirstSector[2 + 2];

	/*** Used for looping. ***/
	int iLoopEntry;
	int iLoopChar;

	LSeek (iFd, 0x1000);
	for (iLoopEntry = 1; iLoopEntry <= ENTRIES; iLoopEntry++)
	{
		ReadFromFile (iFd, "Base Name", 8, sBaseName);
		for (iLoopChar = 7; iLoopChar > 0; iLoopChar--)
			{ if (sBaseName[iLoopChar] == ' ') { sBaseName[iLoopChar] = '\0'; } }
		ReadFromFile (iFd, "Extension", 3, sExtension);
		snprintf (sFileNames[iLoopEntry], MAX_FILENAME,
			"%s.%s", sBaseName, sExtension);
		ReadFromFile (iFd, "", 3, sUnknown);
		ReadFromFile (iFd, "First Sector", 2, sFirstSector);
		iFirstSectors[iLoopEntry] = sFirstSector[0] + (sFirstSector[1] << 8);

		if (iDebug == 1)
		{
			if (iFirstSectors[iLoopEntry] != 65535)
			{
				printf ("[ INFO ] First sector %s: %i\n",
					sFileNames[iLoopEntry], iFirstSectors[iLoopEntry]);
			} else {
				printf ("[ INFO ] (Unused entry.)\n");
			}
		}
	}
}
/*****************************************************************************/
void DecompressFile (void)
/*****************************************************************************/
{
	/*** This function decompresses sFileCom into sFileDecom. ***/

	unsigned char sArg[4 + 2];
	unsigned char cN, cA, cB, cC, cD;
	unsigned char cLast1, cLast2, cLast3, cLast4;
	unsigned char cPos1, cPos2, cPos3, cPos4;
	int iTimes;

	/*** Used for looping. ***/
	int iLoopRepeat;

	sFileDecom[0] = '\0';
	iSizeDecom = 0;

	iPosIn = 0;
	do {
		if (iDebug == 1)
		{
			printf ("[ INFO ] Compression at %i: 0x%02X\n",
				iPosIn, sFileCom[iPosIn]);
		}
		switch (sFileCom[iPosIn])
		{
			case 0x00:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				sArg[4] = sFileCom[iPosIn + 4];
				iPosIn+=5;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[4];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[4]); }
				iSizeDecom+=4;
				break;
			case 0x01:
				/*** No arguments. ***/
				iPosIn+=1;
				/***/
				sFileDecom[iSizeDecom + 0] = sFileDecom[iSizeDecom - 4];
				sFileDecom[iSizeDecom + 1] = sFileDecom[iSizeDecom - 3];
				sFileDecom[iSizeDecom + 2] = sFileDecom[iSizeDecom - 2];
				sFileDecom[iSizeDecom + 3] = sFileDecom[iSizeDecom - 1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sFileDecom[iSizeDecom - 4], sFileDecom[iSizeDecom - 3],
					sFileDecom[iSizeDecom - 2], sFileDecom[iSizeDecom - 1]); }
				iSizeDecom+=4;
				break;
			case 0x11:
				sArg[1] = sFileCom[iPosIn + 1];
				iPosIn+=2;
				/***/
				cLast1 = sFileDecom[iSizeDecom - 4]; //TODO: maybe move into for()
				cLast2 = sFileDecom[iSizeDecom - 3];
				cLast3 = sFileDecom[iSizeDecom - 2];
				cLast4 = sFileDecom[iSizeDecom - 1];
				iTimes = sArg[1] + 1;
				for (iLoopRepeat = 1; iLoopRepeat <= iTimes; iLoopRepeat++)
				{
					sFileDecom[iSizeDecom + 0] = cLast1;
					sFileDecom[iSizeDecom + 1] = cLast2;
					sFileDecom[iSizeDecom + 2] = cLast3;
					sFileDecom[iSizeDecom + 3] = cLast4;
					if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
						cLast1, cLast2, cLast3, cLast4); }
					iSizeDecom+=4;
				}
				break;
			case 0x21:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				cLast1 = sFileDecom[iSizeDecom - 4]; //TODO: maybe move into for()
				cLast2 = sFileDecom[iSizeDecom - 3];
				cLast3 = sFileDecom[iSizeDecom - 2];
				cLast4 = sFileDecom[iSizeDecom - 1];
				iTimes = sArg[1] + (sArg[2] << 8) + 1;
				for (iLoopRepeat = 1; iLoopRepeat <= iTimes; iLoopRepeat++)
				{
					sFileDecom[iSizeDecom + 0] = cLast1;
					sFileDecom[iSizeDecom + 1] = cLast2;
					sFileDecom[iSizeDecom + 2] = cLast3;
					sFileDecom[iSizeDecom + 3] = cLast4;
					if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
						cLast1, cLast2, cLast3, cLast4); }
					iSizeDecom+=4;
				}
				break;
			case 0x81:
				/*** No arguments. ***/
				iPosIn+=1;
				/***/
				sFileDecom[iSizeDecom + 0] = sFileDecom[iSizeDecom - 8];
				sFileDecom[iSizeDecom + 1] = sFileDecom[iSizeDecom - 7];
				sFileDecom[iSizeDecom + 2] = sFileDecom[iSizeDecom - 6];
				sFileDecom[iSizeDecom + 3] = sFileDecom[iSizeDecom - 5];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sFileDecom[iSizeDecom - 8], sFileDecom[iSizeDecom - 7],
					sFileDecom[iSizeDecom - 6], sFileDecom[iSizeDecom - 5]); }
				iSizeDecom+=4;
				break;
			case 0x91:
				sArg[1] = sFileCom[iPosIn + 1];
				iPosIn+=2;
				/***/
				cLast1 = sFileDecom[iSizeDecom - 8]; // TODO: maybe move into for()
				cLast2 = sFileDecom[iSizeDecom - 7];
				cLast3 = sFileDecom[iSizeDecom - 6];
				cLast4 = sFileDecom[iSizeDecom - 5];
				iTimes = sArg[1] + 1;
				for (iLoopRepeat = 1; iLoopRepeat <= iTimes; iLoopRepeat++)
				{
					sFileDecom[iSizeDecom + 0] = cLast1;
					sFileDecom[iSizeDecom + 1] = cLast2;
					sFileDecom[iSizeDecom + 2] = cLast3;
					sFileDecom[iSizeDecom + 3] = cLast4;
					if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
						cLast1, cLast2, cLast3, cLast4); }
					iSizeDecom+=4;
				}
				break;
			case 0x02:
				sArg[1] = sFileCom[iPosIn + 1];
				iPosIn+=2;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[1], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x03:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[2], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x13:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[1], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x23:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[2], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x33:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[1], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x04:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[2], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x14:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[1], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x24:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[2], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x44:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[2], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x54:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[1], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x64:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x74:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[2], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x84:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x94:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x05: case 0x15: case 0x25: case 0x35: case 0x45: case 0x55:
			case 0x65: case 0x75: case 0x85: case 0x95: case 0xA5: case 0xB5:
			case 0xC5: case 0xD5: case 0xE5: case 0xF5:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				cN = NibbleUpper (sFileCom[iPosIn - 3]);
				cA = NibbleLower (sArg[1]);
				cB = NibbleUpper (sArg[1]);
				cC = NibbleLower (sArg[2]);
				cD = NibbleUpper (sArg[2]);
				sFileDecom[iSizeDecom + 0] = NibbleMerge (cN, cA);
				sFileDecom[iSizeDecom + 1] = NibbleMerge (cN, cB);
				sFileDecom[iSizeDecom + 2] = NibbleMerge (cN, cC);
				sFileDecom[iSizeDecom + 3] = NibbleMerge (cN, cD);
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					NibbleMerge (cN, cA), NibbleMerge (cN, cB),
					NibbleMerge (cN, cC), NibbleMerge (cN, cD)); }
				iSizeDecom+=4;
				break;
			case 0x06: case 0x16: case 0x26: case 0x36: case 0x46: case 0x56:
			case 0x66: case 0x76: case 0x86: case 0x96: case 0xA6: case 0xB6:
			case 0xC6: case 0xD6: case 0xE6: case 0xF6:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				cN = NibbleUpper (sFileCom[iPosIn - 3]);
				cA = NibbleLower (sArg[1]);
				cB = NibbleUpper (sArg[1]);
				cC = NibbleLower (sArg[2]);
				cD = NibbleUpper (sArg[2]);
				sFileDecom[iSizeDecom + 0] = NibbleMerge (cA, cN);
				sFileDecom[iSizeDecom + 1] = NibbleMerge (cB, cN);
				sFileDecom[iSizeDecom + 2] = NibbleMerge (cC, cN);
				sFileDecom[iSizeDecom + 3] = NibbleMerge (cD, cN);
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					NibbleMerge (cA, cN), NibbleMerge (cB, cN),
					NibbleMerge (cC, cN), NibbleMerge (cD, cN)); }
				iSizeDecom+=4;
				break;
			case 0x07: case 0x17: case 0x27: case 0x37: case 0x47: case 0x57:
			case 0x67: case 0x77: case 0x87: case 0x97: case 0xA7: case 0xB7:
			case 0xC7: case 0xD7: /*** No 0xE7. ***/ case 0xF7:
				DecompFun (sFileCom[iPosIn] - 7, 0x00, 0x00, 0x00, 0x00);
				break;
			case 0x08: case 0x18: case 0x28: case 0x38: case 0x48: case 0x58:
			case 0x68: case 0x78: case 0x88: case 0x98: case 0xA8: case 0xB8:
			case 0xC8: case 0xD8: /*** No 0xE8. ***/ case 0xF8:
				DecompFun (sFileCom[iPosIn] - 8, 0xFF, 0xFF, 0xFF, 0xFF);
				break;
			case 0x09: case 0x19: case 0x29: case 0x39: case 0x49: case 0x59:
			case 0x69: case 0x79: case 0x89: case 0x99: case 0xA9: case 0xB9:
			case 0xC9: case 0xD9: /*** No 0xE9 or 0xF9. ***/
				cPos1 = sFileDecom[iSizeDecom - 4];
				cPos2 = sFileDecom[iSizeDecom - 3];
				cPos3 = sFileDecom[iSizeDecom - 2];
				cPos4 = sFileDecom[iSizeDecom - 1];
				DecompFun (sFileCom[iPosIn] - 9, cPos1, cPos2,
					cPos3, cPos4);
				break;
			case 0x0A: case 0x1A: case 0x2A: case 0x3A: case 0x4A: case 0x5A:
			case 0x6A: case 0x7A: case 0x8A: case 0x9A: case 0xAA: case 0xBA:
			case 0xCA: case 0xDA: /*** No 0xEA or 0xFA. ***/
				cPos1 = sFileDecom[iSizeDecom - 8];
				cPos2 = sFileDecom[iSizeDecom - 7];
				cPos3 = sFileDecom[iSizeDecom - 6];
				cPos4 = sFileDecom[iSizeDecom - 5];
				DecompFun (sFileCom[iPosIn] - 10, cPos1, cPos2,
					cPos3, cPos4);
				break;
			default:
				printf ("[FAILED] Invalid compression. Head byte at pos. %i/%i:"
					" 0x%02X!\n", iPosIn, iSizeCom, sFileCom[iPosIn]);
				exit (EXIT_ERROR);
				break;
		}
	} while (iPosIn < iSizeCom);
	sFileDecom[iSizeDecom] = '\0';
}
/*****************************************************************************/
void DecompFun (int iType, int cPos1, int cPos2, int cPos3, int cPos4)
/*****************************************************************************/
{
	unsigned char sArg[4 + 2];

	switch (iType)
	{
		case 0x00:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[1];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, cPos3, sArg[1]); }
			iSizeDecom+=4;
			break;
		case 0x10:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[1];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, sArg[1], cPos4); }
			iSizeDecom+=4;
			break;
		case 0x20:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], cPos3, cPos4); }
			iSizeDecom+=4;
			break;
		case 0x30:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, cPos3, cPos4); }
			iSizeDecom+=4;
			break;
		case 0x40:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[1];
			sFileDecom[iSizeDecom + 3] = sArg[2];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, sArg[1], sArg[2]); }
			iSizeDecom+=4;
			break;
		case 0x50:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[2];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], cPos3, sArg[2]); }
			iSizeDecom+=4;
			break;
		case 0x60:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], sArg[2], cPos4); }
			iSizeDecom+=4;
			break;
		case 0x70:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[2];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, cPos3, sArg[2]); }
			iSizeDecom+=4;
			break;
		case 0x80:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, sArg[2], cPos4); }
			iSizeDecom+=4;
			break;
		case 0x90:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = sArg[2];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], sArg[2], cPos3, cPos4); }
			iSizeDecom+=4;
			break;
		case 0xA0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = sArg[3];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], sArg[2], sArg[3]); }
			iSizeDecom+=4;
			break;
		case 0xB0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = sArg[3];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, sArg[2], sArg[3]); }
			iSizeDecom+=4;
			break;
		case 0xC0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = sArg[2];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[3];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], sArg[2], cPos3, sArg[3]); }
			iSizeDecom+=4;
			break;
		case 0xD0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = sArg[2];
			sFileDecom[iSizeDecom + 2] = sArg[3];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], sArg[2], sArg[3], cPos4); }
			iSizeDecom+=4;
			break;
		case 0xF0:
			/*** No arguments. ***/
			iPosIn+=1;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, cPos3, cPos4); }
			iSizeDecom+=4;
			break;
	}
}
/*****************************************************************************/
int CheckHeader (int iFd, int iOffset)
/*****************************************************************************/
{
	unsigned char sCheckHeader[2 + 2];

	LSeek (iFd, iOffset);
	ReadFromFile (iFd, "Check Header", 2, sCheckHeader);
	if ((sCheckHeader[0] == 0xEB) && (sCheckHeader[1] == 0x0A))
		{ return (1); } else { return (0); }
}
/*****************************************************************************/
unsigned char NibbleLower (unsigned char cChar)
/*****************************************************************************/
{
	return (cChar & 0x0f); /*** 0f = 00001111 as mask ***/
}
/*****************************************************************************/
unsigned char NibbleUpper (unsigned char cChar)
/*****************************************************************************/
{
	return ((cChar & 0xf0) >> 4); /*** f0 = 11110000 as mask ***/
}
/*****************************************************************************/
unsigned char NibbleMerge (unsigned char cUpper, unsigned char cLower)
/*****************************************************************************/
{
	return ((cUpper << 4) | cLower);
}
/*****************************************************************************/
void SaveFileDecom (char *sFileName)
/*****************************************************************************/
{
	/*** Function not currently in use. ***/

	int iFdD;

	iFdD = open (sFileName, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0600);

	WriteCharByChar (iFdD, sFileDecom, iSizeDecom);

	close (iFdD);
}
/*****************************************************************************/
void LoadLevel (int iLevel)
/*****************************************************************************/
{
	char sFileName[MAX_FILENAME + 2];
	int iPos;
	int iBlock;
	int iLeft, iRight, iUp, iDown;
	char sName[7 + 2];
	int iFlagObject, iFlag, iObject;

	/*** Used for looping. ***/
	int iLoopBlock;
	int iLoopRoom;
	int iLoopTile;
	int iLoopMinus;

	iChanged = 0; /*** TODO correct location? ***/
	snprintf (sFileName, MAX_FILENAME, "LEV%02i.MAP", iLevel);
	LoadFileCom (iFdA, sFileName);
	DecompressFile();

	iPos = 0;
	// TODO CHR tiles
	iPos+=2048;
	// TODO back layer halfblocks
	iPos+=256;
	// TODO front layer halfblocks
	iPos+=256;
	PrIfDe ("[  OK  ] Loading: Flags and Objects\n");
	for (iLoopBlock = 0; iLoopBlock < BLOCKS; iLoopBlock++)
	{
		iFlagObject = sFileDecom[iPos + (iLoopBlock * 2)];
		iFlag = iFlagObject & 0x07;
		iObject = iFlagObject & 0xf8;
		arBlockFlag[iLoopBlock] = iFlag;
		arBlockObject[iLoopBlock] = iObject;
		if (iDebug == 1)
		{
			printf ("[ INFO ] Block %i: flag %i, object %i\n",
				iLoopBlock, iFlag, iObject);
		}
	}
	iPos+=256;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		if (iDebug == 1) { printf ("\nRoom: %i\n\n", iLoopRoom); }
		for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
		{
			iBlock = sFileDecom[iPos + ((iLoopRoom - 1) * 30) + (iLoopTile - 1)];
			arBlock[iLoopRoom][iLoopTile] = iBlock;
			if (iDebug == 1)
			{
				GetBlockName (iBlock, sName);
				printf ("%s", sName);
				if ((iLoopTile != 10) && (iLoopTile != 20) && (iLoopTile != 30))
					{ printf ("|"); }
				if ((iLoopTile == 10) || (iLoopTile == 20))
				{
					printf ("\n");
					for (iLoopMinus = 1; iLoopMinus <= 79; iLoopMinus++)
						{ printf ("-"); }
					printf ("\n");
				}
			}
		}
		if (iDebug == 1) { printf ("\n\n"); }
	}
	iPos+=720;
	// TODO modifiers in each room
	iPos+=720;
	// TODO door events 1
	iPos+=256;
	// TODO door events 2
	iPos+=256;
	PrIfDe ("[  OK  ] Loading: Room Links\n");
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		iLeft = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 0];
		iRight = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 1];
		iUp = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 2];
		iDown = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 3];
		arRoomLinks[iLoopRoom][1] = iLeft;
		arRoomLinks[iLoopRoom][2] = iRight;
		arRoomLinks[iLoopRoom][3] = iUp;
		arRoomLinks[iLoopRoom][4] = iDown;
		if (iDebug == 1)
		{
			printf ("[ INFO ] Room %i is connected to room (0 = none):"
				" l%i r%i u%i d%i\n", iLoopRoom, iLeft, iRight, iUp, iDown);
		}
	}
	iPos+=96;
	// TODO number of used rooms
	iPos+=1;
	// TODO unused
	iPos+=24;
	// TODO unused
	iPos+=24;
	// TODO unused
	iPos+=15;
	// TODO starting room
	iKidRoom = sFileDecom[iPos];
	iKidPos = sFileDecom[iPos + 1] + 1;
	iKidDir = sFileDecom[iPos + 2];
	if ((iLevel == 1) || (iLevel == 13))
	{
		if (iKidDir == 0x00) { iKidDir = 0xFF; } else { iKidDir = 0x00; }
	}
	if (iDebug == 1)
	{
		printf ("[ INFO ] The kid starts in room: %i, position: %i, turned: "
			"%c\n", iKidRoom, iKidPos, cShowDirection (iKidDir));
	}
	iPos+=3;
	// TODO unused
	iPos+=4;
	// TODO guard pos
	iPos+=24;
	// TODO guard dir
	iPos+=24;
	// TODO unused
	iPos+=24;
	// TODO unused
	iPos+=24;
	// TODO guard skill
	iPos+=24;
	// TODO unused
	iPos+=24;
	// TODO unused
	iPos+=24;
	// TODO unused (17 bytes)

	PrIfDe ("[  OK  ] Checking for broken room links.\n");
	iBrokenRoomLinks = BrokenRoomLinks (1);

	iCurLevel = iLevel;
	iCurRoom = iKidRoom;
	iOnTile = 1;
	iPlaytest = 0;
}
/*****************************************************************************/
void InitScreenAction (char *sAction)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void InitScreen (void)
/*****************************************************************************/
{
	SDL_AudioSpec fmt;
	char sImage[MAX_IMG + 2];
	SDL_Surface *imgicon;
	int iJoyNr;
	SDL_Rect barbox;
	SDL_Event event;
	DIR *dDir;
	struct dirent *stDirent;
	char sBlock[3 + 2]; /*** 2 chars, plus \0 ***/
	int iXPosOld, iYPosOld;
	int iLocX, iLocY;
	char sExtra[20 + 2];
	const Uint8 *keystate;
	Uint32 oldticks, newticks;

	/*** Used for looping. ***/
	int iLoopExtra;
	int iLoopTile;

	if (SDL_Init (SDL_INIT_AUDIO|SDL_INIT_VIDEO|
		SDL_INIT_GAMECONTROLLER|SDL_INIT_HAPTIC) < 0)
	{
		printf ("[FAILED] Unable to init SDL: %s!\n", SDL_GetError());
		exit (EXIT_ERROR);
	}
	atexit (SDL_Quit);

	/*** main window ***/
	window = SDL_CreateWindow (EDITOR_NAME " " EDITOR_VERSION,
		SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
		(WINDOW_WIDTH) * iScale, (WINDOW_HEIGHT) * iScale, iFullscreen);
	if (window == NULL)
	{
		printf ("[FAILED] Unable to create main window: %s!\n", SDL_GetError());
		exit (EXIT_ERROR);
	}
	iWindowID = SDL_GetWindowID (window);
	iActiveWindowID = iWindowID;
	ascreen = SDL_CreateRenderer (window, -1, 0);
	if (ascreen == NULL)
	{
		printf ("[FAILED] Unable to set video mode: %s!\n", SDL_GetError());
		exit (EXIT_ERROR);
	}

	/*** Some people may prefer linear, but we're going old school. ***/
	SDL_SetHint (SDL_HINT_RENDER_SCALE_QUALITY, "nearest");
	if (iFullscreen != 0)
	{
		SDL_RenderSetLogicalSize (ascreen, (WINDOW_WIDTH) * iScale,
			(WINDOW_HEIGHT) * iScale);
	}

	if (TTF_Init() == -1)
	{
		exit (EXIT_ERROR);
	}

	LoadFonts();

	curArrow = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_ARROW);
	curWait = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_WAIT);
	curHand = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_HAND);

	if (iNoAudio != 1)
	{
		PrIfDe ("[  OK  ] Initializing Audio\n");
		fmt.freq = 44100;
		fmt.format = AUDIO_S16;
		fmt.channels = 2;
		fmt.samples = 512;
		fmt.callback = MixAudio;
		fmt.userdata = NULL;
		if (SDL_OpenAudio (&fmt, NULL) < 0)
		{
			printf ("[FAILED] Unable to open audio: %s!\n", SDL_GetError());
			exit (EXIT_ERROR);
		}
		SDL_PauseAudio (0);
	}

	/*** main window icon ***/
	snprintf (sImage, MAX_IMG, "%s%s", PNG_VARIOUS, "TODO_icon.png");
	imgicon = IMG_Load (sImage);
	if (imgicon == NULL)
	{
		printf ("[ WARN ] Could not load \"%s\": %s!\n", sImage, strerror (errno));
	} else {
		SDL_SetWindowIcon (window, imgicon);
	}

	/*** Open the first available controller. ***/
	iController = 0;
	if (iNoController != 1)
	{
		for (iJoyNr = 0; iJoyNr < SDL_NumJoysticks(); iJoyNr++)
		{
			if (SDL_IsGameController (iJoyNr))
			{
				controller = SDL_GameControllerOpen (iJoyNr);
				if (controller)
				{
					snprintf (sControllerName, MAX_CON, "%s",
						SDL_GameControllerName (controller));
					if (iDebug == 1)
					{
						printf ("[ INFO ] Found a controller \"%s\"; \"%s\".\n",
							sControllerName, SDL_GameControllerNameForIndex (iJoyNr));
					}
					joystick = SDL_GameControllerGetJoystick (controller);
					iController = 1;

					/*** Just for fun, use haptic. ***/
					if (SDL_JoystickIsHaptic (joystick))
					{
						haptic = SDL_HapticOpenFromJoystick (joystick);
						if (SDL_HapticRumbleInit (haptic) == 0)
						{
							SDL_HapticRumblePlay (haptic, 1.0, 1000);
						} else {
							printf ("[ WARN ] Could not initialize the haptic device: %s\n",
								SDL_GetError());
						}
					} else {
						PrIfDe ("[ INFO ] The game controller is not haptic.\n");
					}
				} else {
					printf ("[ WARN ] Could not open game controller %i: %s\n",
						iController, SDL_GetError());
				}
			}
		}
		if (iController != 1) { PrIfDe ("[ INFO ] No controller found.\n"); }
	} else {
		PrIfDe ("[ INFO ] Using keyboard and mouse.\n");
	}

	/*******************/
	/* Preload images. */
	/*******************/

	/*** loading ***/
	SDL_SetCursor (curWait);
	PreLoad (PNG_VARIOUS, "loading.png", &imgloading);
	PreLoad (PNG_VARIOUS, "hourglass_sprite.png", &imghourglasssprite);
	PreLoad (PNG_VARIOUS, "sand_sprite.png", &imgsandsprite);
	ShowImage (imgloading, 0, 0, "imgloading", ascreen, iScale, 1);
	SDL_SetRenderDrawColor (ascreen, 0x22, 0x22, 0x22, SDL_ALPHA_OPAQUE);
	barbox.x = 10 * iScale;
	barbox.y = 10 * iScale;
	barbox.w = 20 * iScale;
	barbox.h = 447 * iScale;
	SDL_RenderFillRect (ascreen, &barbox);
	iHourglassFrame = 1;
	iSandFrame = 1;
	SDL_RenderPresent (ascreen);

	iPreLoaded = 0;
	iCurrentBarHeight = 0;
	iNrToPreLoad = 76; /*** Value can be obtained via debug mode. ***/
	SDL_SetCursor (curWait);

	/*** buttons ***/
  PreLoad (PNG_BUTTONS, "left_0.png", &imgleft_0);
  PreLoad (PNG_BUTTONS, "left_1.png", &imgleft_1);
  PreLoad (PNG_BUTTONS, "right_0.png", &imgright_0);
  PreLoad (PNG_BUTTONS, "right_1.png", &imgright_1);
  PreLoad (PNG_BUTTONS, "up_0.png", &imgup_0);
  PreLoad (PNG_BUTTONS, "up_1.png", &imgup_1);
  PreLoad (PNG_BUTTONS, "down_0.png", &imgdown_0);
  PreLoad (PNG_BUTTONS, "down_1.png", &imgdown_1);
  PreLoad (PNG_BUTTONS, "left_right_no.png", &imglrno);
  PreLoad (PNG_BUTTONS, "up_down_no.png", &imgudno);
	if (iController == 0)
	{
		PreLoad (PNG_BUTTONS, "rooms_on_0.png", &imgroomson_0);
		PreLoad (PNG_BUTTONS, "rooms_on_1.png", &imgroomson_1);
		PreLoad (PNG_BUTTONS, "rooms_off.png", &imgroomsoff);
		PreLoad (PNG_BUTTONS, "broken_rooms_on_0.png", &imgbroomson_0);
		PreLoad (PNG_BUTTONS, "broken_rooms_on_1.png", &imgbroomson_1);
		PreLoad (PNG_BUTTONS, "broken_rooms_off.png", &imgbroomsoff);
		PreLoad (PNG_BUTTONS, "up_down_no_nfo.png", &imgudnonfo);
		PreLoad (PNG_BUTTONS, "events_on_0.png", &imgeventson_0);
		PreLoad (PNG_BUTTONS, "events_on_1.png", &imgeventson_1);
		PreLoad (PNG_BUTTONS, "events_off.png", &imgeventsoff);
		PreLoad (PNG_BUTTONS, "save_on_0.png", &imgsaveon_0);
		PreLoad (PNG_BUTTONS, "save_on_1.png", &imgsaveon_1);
		PreLoad (PNG_BUTTONS, "save_off.png", &imgsaveoff);
		PreLoad (PNG_BUTTONS, "quit_0.png", &imgquit_0);
		PreLoad (PNG_BUTTONS, "quit_1.png", &imgquit_1);
		PreLoad (PNG_BUTTONS, "prev_on_0.png", &imgprevon_0);
		PreLoad (PNG_BUTTONS, "prev_on_1.png", &imgprevon_1);
		PreLoad (PNG_BUTTONS, "prev_off.png", &imgprevoff);
		PreLoad (PNG_BUTTONS, "next_on_0.png", &imgnexton_0);
		PreLoad (PNG_BUTTONS, "next_on_1.png", &imgnexton_1);
		PreLoad (PNG_BUTTONS, "next_off.png", &imgnextoff);
	} else {
		PreLoad (PNG_GAMEPAD, "rooms_on_0.png", &imgroomson_0);
		PreLoad (PNG_GAMEPAD, "rooms_on_1.png", &imgroomson_1);
		PreLoad (PNG_GAMEPAD, "rooms_off.png", &imgroomsoff);
		PreLoad (PNG_GAMEPAD, "broken_rooms_on_0.png", &imgbroomson_0);
		PreLoad (PNG_GAMEPAD, "broken_rooms_on_1.png", &imgbroomson_1);
		PreLoad (PNG_GAMEPAD, "broken_rooms_off.png", &imgbroomsoff);
		PreLoad (PNG_GAMEPAD, "up_down_no_nfo.png", &imgudnonfo);
		PreLoad (PNG_GAMEPAD, "events_on_0.png", &imgeventson_0);
		PreLoad (PNG_GAMEPAD, "events_on_1.png", &imgeventson_1);
		PreLoad (PNG_GAMEPAD, "events_off.png", &imgeventsoff);
		PreLoad (PNG_GAMEPAD, "save_on_0.png", &imgsaveon_0);
		PreLoad (PNG_GAMEPAD, "save_on_1.png", &imgsaveon_1);
		PreLoad (PNG_GAMEPAD, "save_off.png", &imgsaveoff);
		PreLoad (PNG_GAMEPAD, "quit_0.png", &imgquit_0);
		PreLoad (PNG_GAMEPAD, "quit_1.png", &imgquit_1);
		PreLoad (PNG_GAMEPAD, "prev_on_0.png", &imgprevon_0);
		PreLoad (PNG_GAMEPAD, "prev_on_1.png", &imgprevon_1);
		PreLoad (PNG_GAMEPAD, "prev_off.png", &imgprevoff);
		PreLoad (PNG_GAMEPAD, "next_on_0.png", &imgnexton_0);
		PreLoad (PNG_GAMEPAD, "next_on_1.png", &imgnexton_1);
		PreLoad (PNG_GAMEPAD, "next_off.png", &imgnextoff);
	}

	/*** living ***/
	PreLoad (PNG_LIVING, "kid_r.png", &imgkidr[1]);
	PreLoad (PNG_LIVING, "kid_l.png", &imgkidl[1]);

	/*** sliving ***/
	PreLoad (PNG_SLIVING, "kid_r.png", &imgkidr[2]);
	PreLoad (PNG_SLIVING, "kid_l.png", &imgkidl[2]);

	/*** various ***/
	PreLoad (PNG_VARIOUS, "black.png", &imgblack);
	PreLoad (PNG_VARIOUS, "unknown.png", &imgunk);
	PreLoad (PNG_VARIOUS, "playtest.png", &imgplaytest);
	PreLoad (PNG_VARIOUS, "sel_block.png", &imgselblock);
	if (iController == 0)
	{
		PreLoad (PNG_VARIOUS, "level_bar.png", &imgbar);
	} else {
		PreLoad (PNG_GAMEPAD, "level_bar.png", &imgbar);
	}

	dDir = opendir (PNG_L0003);
	if (dDir == NULL)
	{
		printf ("[FAILED] Cannot open directory \"%s\": %s!\n",
			PNG_L0003, strerror (errno));
		exit (EXIT_ERROR);
	}
	while ((stDirent = readdir (dDir)) != NULL)
	{
		if ((strcmp (stDirent->d_name, ".") != 0) &&
			(strcmp (stDirent->d_name, "..") != 0))
		{
			snprintf (sBlock, 3, "%c%c",
				stDirent->d_name[0], stDirent->d_name[1]);
			PreLoad (PNG_L0003, stDirent->d_name,
				&imgl0003[(int)strtol(sBlock, NULL, 16)]);
		}
	}
	closedir (dDir);

	/*** extras ***/
	for (iLoopExtra = 0; iLoopExtra <= 10; iLoopExtra++)
	{
		snprintf (sExtra, 20, "extras_%02i.png", iLoopExtra);
		PreLoad (PNG_EXTRAS, sExtra, &extras[iLoopExtra]);
	}

	/*** One final update of the bar. ***/
	LoadingBar (BAR_FULL);

	if (iDebug == 1)
		{ printf ("[ INFO ] Preloaded images: %i\n", iPreLoaded); }
	SDL_SetCursor (curArrow);

	iCurLevel = iStartLevel;
	if (iDebug == 1)
		{ printf ("[ INFO ] Starting in level: %i\n", iStartLevel); }

	iSelected = 1; /*** Start with the upper left selected. ***/
	iScreen = 1;
	iDownAt = 0;
	iChangeEvent = 0;
	oldticks = 0;

	ShowScreen (iScreen);
	InitPopUp();
	while (1)
	{
		if (iNoAnim == 0)
		{
			/*** This is for the animation; 12 fps (1000/83) is fine for PoP. ***/
			newticks = SDL_GetTicks();
			if (newticks > oldticks + 83)
			{
				/*** ... ***/
				ShowScreen (iScreen);
				oldticks = newticks;
			}
		}

		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN: /*** https://wiki.libsdl.org/SDL_Keycode ***/
					switch (event.key.keysym.sym)
					{
						case SDLK_LEFT:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								switch (iScreen)
								{
									case 1:
										if (arRoomLinks[iCurRoom][1] != 0)
										{
											iCurRoom = arRoomLinks[iCurRoom][1];
											PlaySound ("wav/scroll.wav");
										}
										break;
									case 3:
										ChangeEvent (-1, 0);
										break;
								}
							} else if ((event.key.keysym.mod & KMOD_LCTRL) ||
								(event.key.keysym.mod & KMOD_RCTRL))
							{
								if (iScreen == 3) { ChangeEvent (-10, 0); }
							} else {
								InitScreenAction ("left");
							}
							break;
						case SDLK_RIGHT:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								switch (iScreen)
								{
									case 1:
										if (arRoomLinks[iCurRoom][2] != 0)
										{
											iCurRoom = arRoomLinks[iCurRoom][2];
											PlaySound ("wav/scroll.wav");
										}
										break;
									case 3:
										ChangeEvent (1, 0);
										break;
								}
							} else if ((event.key.keysym.mod & KMOD_LCTRL) ||
								(event.key.keysym.mod & KMOD_RCTRL))
							{
								if (iScreen == 3) { ChangeEvent (10, 0); }
							} else {
								InitScreenAction ("right");
							}
							break;
						case SDLK_UP:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								if (iScreen == 1)
								{
									if (arRoomLinks[iCurRoom][3] != 0)
									{
										iCurRoom = arRoomLinks[iCurRoom][3];
										PlaySound ("wav/scroll.wav");
									}
								}
							} else {
								InitScreenAction ("up");
							}
							break;
						case SDLK_DOWN:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								if (iScreen == 1)
								{
									if (arRoomLinks[iCurRoom][4] != 0)
									{
										iCurRoom = arRoomLinks[iCurRoom][4];
										PlaySound ("wav/scroll.wav");
									}
								}
							} else {
								InitScreenAction ("down");
							}
							break;
						case SDLK_MINUS:
						case SDLK_KP_MINUS:
							if ((iChanged != 0) && (iCurLevel != 1)) { InitPopUpSave(); }
							Prev(); break;
						case SDLK_KP_PLUS:
						case SDLK_EQUALS:
							if ((iChanged != 0) && (iCurLevel != 0)) { InitPopUpSave(); }
							Next(); break;
						case SDLK_ESCAPE:
						case SDLK_q:
							switch (iScreen)
							{
								case 1:
									Quit(); break;
								case 2:
									/*** TODO: iBrokenRoomLinks = BrokenRoomLinks (0); ***/
									/*** no break ***/
								case 3:
									iScreen = 1; break;
							}
							break;
					}
					ShowScreen (iScreen);
					break;
				case SDL_MOUSEMOTION:
					iXPosOld = iXPos;
					iYPosOld = iYPos;
					iXPos = event.motion.x;
					iYPos = event.motion.y;
					if ((iXPosOld == iXPos) && (iYPosOld == iYPos)) { break; }

					if (OnLevelBar() == 1)
						{ iPlaytest = 1; } else { iPlaytest = 0; }

					if (iScreen == 1)
					{
						/*** hover ***/
						for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
						{
							if ((iLoopTile >= 1) && (iLoopTile <= 10))
								{ iLocX = iLoopTile; iLocY = 1; }
							if ((iLoopTile >= 11) && (iLoopTile <= 20))
								{ iLocX = iLoopTile - 10; iLocY = 2; }
							if ((iLoopTile >= 21) && (iLoopTile <= 30))
								{ iLocX = iLoopTile - 20; iLocY = 3; }

							if ((InArea (25 + (64 * (iLocX - 1)), 58 + (128 * (iLocY - 1)),
								25 + (64 * iLocX), 58 + (128 * iLocY)) == 1) &&
								(iSelected != iLoopTile)) { iSelected = iLoopTile; }
						}

						/*** extras ***/
						if (InArea (608, 3, 608 + 49, 3 + 19) == 0) { iExtras = 0; }
						if (InArea (608, 3, 608 + 9, 3 + 9) == 1) { iExtras = 1; }
						if (InArea (618, 3, 618 + 9, 3 + 9) == 1) { iExtras = 2; }
						if (InArea (628, 3, 628 + 9, 3 + 9) == 1) { iExtras = 3; }
						if (InArea (638, 3, 638 + 9, 3 + 9) == 1) { iExtras = 4; }
						if (InArea (648, 3, 648 + 9, 3 + 9) == 1) { iExtras = 5; }
						if (InArea (608, 13, 608 + 9, 13 + 9) == 1) { iExtras = 6; }
						if (InArea (618, 13, 618 + 9, 13 + 9) == 1) { iExtras = 7; }
						if (InArea (628, 13, 628 + 9, 13 + 9) == 1) { iExtras = 8; }
						if (InArea (638, 13, 638 + 9, 13 + 9) == 1) { iExtras = 9; }
						if (InArea (648, 13, 648 + 9, 13 + 9) == 1) { iExtras = 10; }
					}

					// TODO

					ShowScreen (iScreen);
					break;
				case SDL_MOUSEBUTTONDOWN:
					if (event.button.button == 1)
					{
						if (InArea (0, 50, 0 + 25, 50 + 392) == 1) /*** left ***/
						{
							if (arRoomLinks[iCurRoom][1] != 0) { iDownAt = 1; }
						}
						if (InArea (665, 50, 665 + 25, 50 + 392) == 1) /*** right ***/
						{
							if (arRoomLinks[iCurRoom][2] != 0) { iDownAt = 2; }
						}
						if (InArea (25, 25, 25 + 640, 25 + 25) == 1) /*** up ***/
						{
							if (arRoomLinks[iCurRoom][3] != 0) { iDownAt = 3; }
						}
						if (InArea (25, 442, 25 + 640, 442 + 25) == 1) /*** down ***/
						{
							if (arRoomLinks[iCurRoom][4] != 0) { iDownAt = 4; }
						}
						if (InArea (0, 25, 0 + 25, 25 + 25) == 1) /*** rooms ***/
						{
							if (iBrokenRoomLinks == 0)
								{ iDownAt = 5; } else { iDownAt = 11; }
						}
						if (InArea (665, 25, 665 + 25, 25 + 25) == 1) /*** events ***/
							{ iDownAt = 6; }
						if (InArea (0, 442, 0 + 25, 442 + 25) == 1) /*** save ***/
							{ iDownAt = 7; }
						if (InArea (665, 442, 665 + 25, 442 + 25) == 1) /*** quit ***/
							{ iDownAt = 8; }
						if (InArea (0, 0, 0 + 25, 0 + 25) == 1) /*** prev level ***/
							{ iDownAt = 9; }
						if (InArea (665, 0, 665 + 25, 0 + 25) == 1) /*** next level ***/
							{ iDownAt = 10; }
						// TODO
					}
					// TODO
					ShowScreen (iScreen);
					break;
				case SDL_MOUSEBUTTONUP:
					iDownAt = 0;
					if (event.button.button == 1)
					{
						if (InArea (0, 50, 0 + 25, 50 + 392) == 1) /*** left ***/
						{
							if (arRoomLinks[iCurRoom][1] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][1];
								PlaySound ("wav/scroll.wav");
							}
						}
						if (InArea (665, 50, 665 + 25, 50 + 392) == 1) /*** right ***/
						{
							if (arRoomLinks[iCurRoom][2] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][2];
								PlaySound ("wav/scroll.wav");
							}
						}
						if (InArea (25, 25, 25 + 640, 25 + 25) == 1) /*** up ***/
						{
							if (arRoomLinks[iCurRoom][3] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][3];
								PlaySound ("wav/scroll.wav");
							}
						}
						if (InArea (25, 442, 25 + 640, 442 + 25) == 1) /*** down ***/
						{
							if (arRoomLinks[iCurRoom][4] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][4];
								PlaySound ("wav/scroll.wav");
							}
						}
						// TODO
						if (InArea (665, 442, 665 + 25, 442 + 25) == 1) /*** quit ***/
						{
							switch (iScreen)
							{
								case 1:
									Quit(); break;
								case 2:
									iBrokenRoomLinks = BrokenRoomLinks (0);
									iScreen = 1; break;
								case 3:
									iScreen = 1; break;
							}
						}
						if (InArea (0, 0, 0 + 25, 0 + 25) == 1) /*** prev level ***/
						{
							if ((iChanged != 0) && (iCurLevel != 1)) { InitPopUpSave(); }
							Prev();
							break; /*** Stop processing SDL_MOUSEBUTTONUP. ***/
						}
						if (InArea (665, 0, 665 + 25, 0 + 25) == 1) /*** next level ***/
						{
							if ((iChanged != 0) && (iCurLevel != 0)) { InitPopUpSave(); }
							Next();
							break; /*** Stop processing SDL_MOUSEBUTTONUP. ***/
						}
						// TODO
					}
					// TODO
					ShowScreen (iScreen);
					break;
				case SDL_MOUSEWHEEL:
					if (event.wheel.y > 0) /*** scroll wheel up ***/
					{
						if (InArea (25, 50, 25 + 640, 50 + 392) == 1) /*** tiles ***/
						{
							keystate = SDL_GetKeyboardState (NULL);
							if ((keystate[SDL_SCANCODE_LSHIFT]) ||
								(keystate[SDL_SCANCODE_RSHIFT]))
							{ /*** right ***/
								if (arRoomLinks[iCurRoom][2] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][2];
									PlaySound ("wav/scroll.wav");
								}
							} else { /*** up ***/
								if (arRoomLinks[iCurRoom][3] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][3];
									PlaySound ("wav/scroll.wav");
								}
							}
						}
					}
					if (event.wheel.y < 0) /*** scroll wheel down ***/
					{
						if (InArea (25, 50, 25 + 640, 50 + 392) == 1) /*** tiles ***/
						{
							keystate = SDL_GetKeyboardState (NULL);
							if ((keystate[SDL_SCANCODE_LSHIFT]) ||
								(keystate[SDL_SCANCODE_RSHIFT]))
							{ /*** left ***/
								if (arRoomLinks[iCurRoom][1] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][1];
									PlaySound ("wav/scroll.wav");
								}
							} else { /*** down ***/
								if (arRoomLinks[iCurRoom][4] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][4];
									PlaySound ("wav/scroll.wav");
								}
							}
						}
					}
					ShowScreen (iScreen);
					break;
				case SDL_WINDOWEVENT:
					switch (event.window.event)
					{
						case SDL_WINDOWEVENT_EXPOSED:
							ShowScreen (iScreen); break;
						case SDL_WINDOWEVENT_CLOSE:
							Quit(); break;
						case SDL_WINDOWEVENT_FOCUS_GAINED:
							iActiveWindowID = iWindowID; break;
					}
					break;
				case SDL_QUIT:
					Quit(); break;
			}
		}

		/*** prevent CPU eating ***/
		gamespeed = REFRESH;
		while ((SDL_GetTicks() - looptime) < gamespeed)
		{
			SDL_Delay (10);
		}
		looptime = SDL_GetTicks();
	}
}
/*****************************************************************************/
void Quit (void)
/*****************************************************************************/
{
	if (iChanged != 0) { InitPopUpSave(); }
	TTF_CloseFont (font11);
	TTF_CloseFont (font15);
	TTF_CloseFont (font20);
	TTF_Quit();
	SDL_Quit();
	exit (EXIT_NORMAL);
}
/*****************************************************************************/
void GetBlockName (int iBlock, char *sName)
/*****************************************************************************/
{
	switch (iBlock)
	{
		case 0x00: snprintf (sName, 9, "%s", "WallMid"); break;
		case 0x02: snprintf (sName, 9, "%s", "WallLft"); break;
		case 0x05: snprintf (sName, 9, "%s", "Floor05"); break;
		case 0x07: snprintf (sName, 9, "%s", "Floor07"); break;
		case 0x2D: snprintf (sName, 9, "%s", "Floor2D"); break;
		case 0x35: snprintf (sName, 9, "%s", "Empty35"); break;
		case 0x39: snprintf (sName, 9, "%s", "Empty39"); break;
		case 0x3A: snprintf (sName, 9, "%s", "GateRgt"); break;
		default: snprintf (sName, 9, "%07X", iBlock); break;
	}
}
/*****************************************************************************/
void GetFlagName (int iFlag, char *sName)
/*****************************************************************************/
{
	switch (iFlag)
	{
		case 0x00: snprintf (sName, 7, "%s", "empty"); break;
		case 0x01: snprintf (sName, 7, "%s", "floor"); break;
		case 0x02: snprintf (sName, 7, "%s", "wall "); break;
		case 0x03: snprintf (sName, 7, "%s", "fl+wl"); break;
		case 0x04: snprintf (sName, 7, "%s", "torch"); break;
		case 0x05: snprintf (sName, 7, "%s", "fl+tr"); break;
		case 0x06: snprintf (sName, 7, "%s", "wl+tr"); break;
		case 0x07: snprintf (sName, 7, "%s", "all  "); break;
		default: snprintf (sName, 7, "%05X", iFlag); break;
	}
}
/*****************************************************************************/
void GetObjectName (int iObject, char *sName)
/*****************************************************************************/
{
	switch (iObject)
	{
		case 0x00: snprintf (sName, 7, "%s", ""); break;
		case 0x08: snprintf (sName, 7, "%s", "loose"); break;
		case 0x10: snprintf (sName, 7, "%s", "raise"); break;
		case 0x18: snprintf (sName, 7, "%s", "drop "); break;
		case 0x20: snprintf (sName, 7, "%s", "gate "); break;
		case 0x28: snprintf (sName, 7, "%s", "spike"); break;
		case 0x30: snprintf (sName, 7, "%s", "heal "); break;
		case 0x38: snprintf (sName, 7, "%s", "hurt "); break;
		case 0x40: snprintf (sName, 7, "%s", "life "); break;
		case 0x48: snprintf (sName, 7, "%s", "flip "); break;
		case 0x50: snprintf (sName, 7, "%s", "float"); break;
		case 0x58: snprintf (sName, 7, "%s", "chomp"); break;
		case 0x60: snprintf (sName, 7, "%s", "sword"); break;
		case 0x68: snprintf (sName, 7, "%s", "mirr "); break;
		case 0x70: snprintf (sName, 7, "%s", "lexit"); break;
		case 0x78: snprintf (sName, 7, "%s", "skel "); break;
		case 0x80: snprintf (sName, 7, "%s", "TODO "); break;
		case 0x88: snprintf (sName, 7, "%s", "stars"); break;
		default: snprintf (sName, 7, "%05X", iObject); break;
	}
}
/*****************************************************************************/
char cShowDirection (int iDirection)
/*****************************************************************************/
{
  switch (iDirection)
	{
		case 0x00: return ('r'); break;
		case 0xFF: return ('l'); break;
		default: return ('?'); break;
	}
}
/*****************************************************************************/
void LoadFonts (void)
/*****************************************************************************/
{
	font11 = TTF_OpenFont ("ttf/Bitstream-Vera-Sans-Bold.ttf", 11 * iScale);
	font15 = TTF_OpenFont ("ttf/Bitstream-Vera-Sans-Bold.ttf", 15 * iScale);
	font20 = TTF_OpenFont ("ttf/Bitstream-Vera-Sans-Bold.ttf", 20 * iScale);
	if ((font11 == NULL) || (font15 == NULL) || (font20 == NULL))
	{
		printf ("[FAILED] TTF_OpenFont() failed!\n");
		exit (EXIT_ERROR);
	}
}
/*****************************************************************************/
void MixAudio (void *unused, Uint8 *stream, int iLen)
/*****************************************************************************/
{
	int iTemp;
	int iAmount;

	if (unused != NULL) { } /*** To prevent warnings. ***/

	SDL_memset (stream, 0, iLen); /*** SDL2 ***/
	for (iTemp = 0; iTemp < NUM_SOUNDS; iTemp++)
	{
		iAmount = (sounds[iTemp].dlen-sounds[iTemp].dpos);
		if (iAmount > iLen)
		{
			iAmount = iLen;
		}
		SDL_MixAudio (stream, &sounds[iTemp].data[sounds[iTemp].dpos], iAmount,
			SDL_MIX_MAXVOLUME);
		sounds[iTemp].dpos += iAmount;
	}
}
/*****************************************************************************/
void PreLoad (char *sPath, char *sPNG, SDL_Texture **imgImage)
/*****************************************************************************/
{
	char sImage[MAX_IMG + 2];
	int iBarHeight;

	snprintf (sImage, MAX_IMG, "%s%s", sPath, sPNG);
	*imgImage = IMG_LoadTexture (ascreen, sImage);
	if (!*imgImage)
	{
		printf ("[FAILED] IMG_LoadTexture: %s!\n", IMG_GetError());
		exit (EXIT_ERROR);
	}

	iPreLoaded++;
	iBarHeight = (int)(((float)iPreLoaded/(float)iNrToPreLoad) * BAR_FULL);
	if (iBarHeight >= iCurrentBarHeight + 10) { LoadingBar (iBarHeight); }
}
/*****************************************************************************/
void ShowImage (SDL_Texture *img, int iX, int iY, char *sImageInfo,
	SDL_Renderer *screen, float fMultiply, int iXYScale)
/*****************************************************************************/
{
	/* Usually, fMultiply is either iScale or ZoomGet().
	 *
	 * Also, usually, iXYScale is 1 with iScale and 0 with ZoomGet().
	 *
	 * If you use this function, make sure to verify the image is properly
	 * (re)scaled and (re)positioned when the main window is zoomed ("z")!
	 */

	SDL_Rect dest;
	SDL_Rect loc;
	int iWidth, iHeight;

	SDL_QueryTexture (img, NULL, NULL, &iWidth, &iHeight);
	loc.x = 0; loc.y = 0; loc.w = iWidth; loc.h = iHeight;
	if (iXYScale == 0)
	{
		dest.x = iX;
		dest.y = iY;
	} else {
		dest.x = iX * fMultiply;
		dest.y = iY * fMultiply;
	}
	dest.w = iWidth * fMultiply;
	dest.h = iHeight * fMultiply;

	/*** This is for the animation. ***/
	if (strcmp (sImageInfo, "imghourglasssprite") == 0)
	{
		loc.x = (iHourglassFrame - 1) * 38;
		loc.w = loc.w / 7;
		dest.w = dest.w / 7;
	}
	if (strcmp (sImageInfo, "imgsandsprite") == 0)
	{
		loc.x = (iSandFrame - 1) * 5;
		loc.w = loc.w / 3;
		dest.w = dest.w / 3;
	}
	if (strcmp (sImageInfo, "imgstatusbarsprite") == 0)
	{
		loc.x = (iStatusBarFrame - 1) * 20;
		loc.w = loc.w / 18;
		dest.w = dest.w / 18;
	}

	if (SDL_RenderCopy (screen, img, &loc, &dest) != 0)
	{
		printf ("[ WARN ] SDL_RenderCopy (%s): %s\n", sImageInfo, SDL_GetError());
	}
}
/*****************************************************************************/
void LoadingBar (int iBarHeight)
/*****************************************************************************/
{
	SDL_Rect bar;

	/*** bar ***/
	bar.x = 12 * iScale;
	bar.y = (455 - iBarHeight) * iScale;
	bar.w = 16 * iScale;
	bar.h = iBarHeight * iScale;
	SDL_SetRenderDrawColor (ascreen, 0x44, 0x44, 0x44, SDL_ALPHA_OPAQUE);
	SDL_RenderFillRect (ascreen, &bar);
	iCurrentBarHeight = iBarHeight;

	/*** hourglass sprite ***/
	iHourglassFrame = floor (((float)iBarHeight/BAR_FULL) * 7);
	ShowImage (imghourglasssprite, 326, 178, "imghourglasssprite",
		ascreen, iScale, 1);

	/*** sand sprite ***/
	if (iBarHeight != BAR_FULL)
	{
		ShowImage (imgsandsprite, 342, 206, "imgsandsprite",
		ascreen, iScale, 1);
		iSandFrame++;
		if (iSandFrame == 4) { iSandFrame = 1; }
	}

	/*** refresh screen ***/
	SDL_RenderPresent (ascreen);
}
/*****************************************************************************/
void ShowScreen (int iScreenS)
/*****************************************************************************/
{
	char sLevel[MAX_TEXT + 2];
	char arText[1 + 2][MAX_TEXT + 2];
	int iImgX, iImgY;
	int iXY[2 + 2];

	/*** Used for looping. ***/
	int iLoopTile;

	/*** black background ***/
	ShowImage (imgblack, 0, 0, "imgblack", ascreen, iScale, 1);

	if (iScreenS == 1)
	{
		for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
		{
			LocationToXY (iLoopTile, iXY);

			ShowTile (arBlock[iCurRoom][iLoopTile], iLoopTile);
			if (iSelected == iLoopTile)
			{
				ShowImage (imgselblock, iXY[0], iXY[1], "imgselblock",
					ascreen, iScale, 1);
			}
		}

		for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
		{
			LocationToXY (iLoopTile, iXY);

			/*** prince ***/
			if ((iCurRoom == iKidRoom) && (iLoopTile == iKidPos))
			{
				switch (iKidDir)
				{
					case 0x00: /*** looks right ***/
						iImgX = iXY[0] + 67;
						iImgY = iXY[1] + 19;
						ShowImage (imgkidr[1], iImgX, iImgY, "imgkidr[1]",
							ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{ ShowImage (imgkidr[2], iImgX, iImgY, "imgkidr[2]",
							ascreen, iScale, 1); }
						break;
					case 0xFF: /*** looks left ***/
						iImgX = iXY[0] + 33;
						iImgY = iXY[1] + 19;
						ShowImage (imgkidl[1], iImgX, iImgY, "imgkidl[1]",
							ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{ ShowImage (imgkidl[2], iImgX, iImgY, "imgkidl[2]",
							ascreen, iScale, 1); }
						break;
					default: printf ("[FAILED] Wrong dir.!\n"); exit (EXIT_ERROR); break;
				}
			}
		}
	}

	if (arRoomLinks[iCurRoom][1] != 0)
	{ /*** left ***/
		if (iDownAt == 1)
		{ /*** down ***/
			ShowImage (imgleft_1, 0, 50, "imgleft_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgleft_0, 0, 50, "imgleft_0", ascreen, iScale, 1);
		}
	} else { /*** no left ***/
		ShowImage (imglrno, 0, 50, "imglrno", ascreen, iScale, 1);
	}
	if (arRoomLinks[iCurRoom][2] != 0)
	{ /*** right ***/
		if (iDownAt == 2)
		{ /*** down ***/
			ShowImage (imgright_1, 665, 50, "imgright_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgright_0, 665, 50, "imgright_0", ascreen, iScale, 1);
		}
	} else { /*** no right ***/
		ShowImage (imglrno, 665, 50, "imglrno", ascreen, iScale, 1);
	}
	if (arRoomLinks[iCurRoom][3] != 0)
	{ /*** up ***/
		if (iDownAt == 3)
		{ /*** down ***/
			ShowImage (imgup_1, 25, 25, "imgup_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgup_0, 25, 25, "imgup_0", ascreen, iScale, 1);
		}
	} else { /*** no up ***/
		if (iScreen != 1)
		{ /*** without info ***/
			ShowImage (imgudno, 25, 25, "imgudno", ascreen, iScale, 1);
		} else { /*** with info ***/
			ShowImage (imgudnonfo, 25, 25, "imgudnonfo", ascreen, iScale, 1);
		}
	}
	if (arRoomLinks[iCurRoom][4] != 0)
	{ /*** down ***/
		if (iDownAt == 4)
		{ /*** down ***/
			ShowImage (imgdown_1, 25, 442, "imgdown_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgdown_0, 25, 442, "imgdown_0", ascreen, iScale, 1);
		}
	} else { /*** no down ***/
		ShowImage (imgudno, 25, 442, "imgudno", ascreen, iScale, 1);
	}

	/*** room links ***/
	if (iScreen != 2)
	{
		if (iBrokenRoomLinks == 1)
		{
			/*** broken room links ***/
			if (iDownAt == 11)
			{ /*** down ***/
				ShowImage (imgbroomson_1, 0, 25, "imgbroomson_1", ascreen, iScale, 1);
			} else { /*** up ***/
				ShowImage (imgbroomson_0, 0, 25, "imgbroomson_0", ascreen, iScale, 1);
			}
		} else {
			/*** regular room links ***/
			if (iDownAt == 5)
			{ /*** down ***/
				ShowImage (imgroomson_1, 0, 25, "imgroomson_1", ascreen, iScale, 1);
			} else { /*** up ***/
				ShowImage (imgroomson_0, 0, 25, "imgroomson_0", ascreen, iScale, 1);
			}
		}
	} else {
		if (iBrokenRoomLinks == 1)
		{
			/*** broken room links ***/
			ShowImage (imgbroomsoff, 0, 25, "imgbroomsoff", ascreen, iScale, 1);
		} else {
			/*** regular room links ***/
			ShowImage (imgroomsoff, 0, 25, "imgroomsoff", ascreen, iScale, 1);
		}
	}

	/*** events ***/
	if (iScreen != 3)
	{
		if (iDownAt == 6)
		{ /*** down ***/
			ShowImage (imgeventson_1, 665, 25, "imgeventson_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgeventson_0, 665, 25, "imgeventson_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgeventsoff, 665, 25, "imgeventsoff", ascreen, iScale, 1);
	}

	/*** save ***/
	if (iChanged != 0)
	{ /*** on ***/
		if (iDownAt == 7)
		{ /*** down ***/
			ShowImage (imgsaveon_1, 0, 442, "imgsaveon_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgsaveon_0, 0, 442, "imgsaveon_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgsaveoff, 0, 442, "imgsaveoff", ascreen, iScale, 1);
	}

	/*** quit ***/
	if (iDownAt == 8)
	{ /*** down ***/
		ShowImage (imgquit_1, 665, 442, "imgquit_1", ascreen, iScale, 1);
	} else { /*** up ***/
		ShowImage (imgquit_0, 665, 442, "imgquit_0", ascreen, iScale, 1);
	}

	/*** prev ***/
	if (iCurLevel != 1)
	{ /*** on ***/
		if (iDownAt == 9)
		{ /*** down ***/
			ShowImage (imgprevon_1, 0, 0, "imgprevon_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgprevon_0, 0, 0, "imgprevon_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgprevoff, 0, 0, "imgprevoff", ascreen, iScale, 1);
	}

	/*** next ***/
	if (iCurLevel != 0)
	{ /*** on ***/
		if (iDownAt == 10)
		{ /*** down ***/
			ShowImage (imgnexton_1, 665, 0, "imgnexton_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgnexton_0, 665, 0, "imgnexton_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgnextoff, 665, 0, "imgnextoff", ascreen, iScale, 1);
	}

	/*** level bar ***/
	ShowImage (imgbar, 25, 0, "imgbar", ascreen, iScale, 1);
	switch (iCurLevel)
	{
		case 0: snprintf (sLevel, MAX_TEXT, "%s", "level 0 (demo)"); break;
		case 1: snprintf (sLevel, MAX_TEXT, "%s", "level 1 (prison)"); break;
		case 2: snprintf (sLevel, MAX_TEXT, "%s", "level 2 (guards)"); break;
		case 3: snprintf (sLevel, MAX_TEXT, "%s", "level 3 (skeleton)"); break;
		case 4: snprintf (sLevel, MAX_TEXT, "%s", "level 4 (mirror)"); break;
		case 5: snprintf (sLevel, MAX_TEXT, "%s", "level 5 (thief)"); break;
		case 6: snprintf (sLevel, MAX_TEXT, "%s", "level 6 (plunge)"); break;
		case 7: snprintf (sLevel, MAX_TEXT, "%s", "level 7 (weightless)"); break;
		case 8: snprintf (sLevel, MAX_TEXT, "%s", "level 8 (mouse)"); break;
		case 9: snprintf (sLevel, MAX_TEXT, "%s", "level 9 (twisty)"); break;
		case 10: snprintf (sLevel, MAX_TEXT, "%s", "level 10 (quad)"); break;
		case 11: snprintf (sLevel, MAX_TEXT, "%s", "level 11 (fragile)"); break;
		case 12: snprintf (sLevel, MAX_TEXT, "%s", "level 12 (12a; tower)"); break;
		case 13: snprintf (sLevel, MAX_TEXT, "%s",
			"level 13 (12b; jaffar)"); break;
		case 14: snprintf (sLevel, MAX_TEXT, "%s", "level 14 (rescue)"); break;
		case 15: snprintf (sLevel, MAX_TEXT, "%s", "level 15 (potions)"); break;
	}
	switch (iScreen)
	{
		case 1:
			snprintf (arText[0], MAX_TEXT, "%s, room %i", sLevel, iCurRoom);
			ShowImage (extras[iExtras], 608, 3, "extras[iExtras]",
				ascreen, iScale, 1);
			break;
		case 2:
			snprintf (arText[0], MAX_TEXT, "%s, room links", sLevel); break;
		case 3:
			snprintf (arText[0], MAX_TEXT, "%s, events", sLevel); break;
	}
	DisplayText (31, 5, 15, arText, 1, font15, color_wh, color_bl);
	if (iPlaytest == 1)
		{ ShowImage (imgplaytest, 25, 50, "imgplaytest", ascreen, iScale, 1); }

	/*** refresh screen ***/
	SDL_RenderPresent (ascreen);
}
/*****************************************************************************/
void InitPopUp (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void InitPopUpSave (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void ShowTile (int iBlock, int iLocation)
/*****************************************************************************/
{
	int iXY[2 + 2];
	SDL_Texture *imgshow;
	char arText[1 + 2][MAX_TEXT + 2];
	/***/
	int iFlag, iObject;
	char sFlagName[5 + 2];
	char sObjectName[5 + 2];

	LocationToXY (iLocation, iXY);

	imgshow = imgunk;
	if ((iCurLevel >= 0) && (iCurLevel <= 3))
	{
		if (imgl0003[iBlock] != NULL)
		{
			imgshow = imgl0003[iBlock];
		}
	}
	// TODO other levels (l0406 l0709 l1011 l1213 l14 l15)
	ShowImage (imgshow, iXY[0], iXY[1], "imgshow", ascreen, iScale, 1);

	iFlag = arBlockFlag[iBlock];
	iObject = arBlockObject[iBlock];
	if ((iFlag != 0) || (iObject != 0))
	{
		GetFlagName (iFlag, sFlagName);
		GetObjectName (iObject, sObjectName);
		snprintf (arText[0], MAX_TEXT, "%s", sFlagName);
		snprintf (arText[1], MAX_TEXT, "%s", sObjectName);
		DisplayText (iXY[0] + 3, iXY[1] + 97, 11, arText,
			2, font11, color_wh, color_bl);
	}
}
/*****************************************************************************/
void LocationToXY (int iLocation, int iXY[4])
/*****************************************************************************/
{
	switch (iLocation)
	{
		/*** above this room ***/
		// TODO
		/*** this room ***/
		case 1: iXY[0] = 25+(64*0); iXY[1] = 58+(128*0); break;
		case 2: iXY[0] = 25+(64*1); iXY[1] = 58+(128*0); break;
		case 3: iXY[0] = 25+(64*2); iXY[1] = 58+(128*0); break;
		case 4: iXY[0] = 25+(64*3); iXY[1] = 58+(128*0); break;
		case 5: iXY[0] = 25+(64*4); iXY[1] = 58+(128*0); break;
		case 6: iXY[0] = 25+(64*5); iXY[1] = 58+(128*0); break;
		case 7: iXY[0] = 25+(64*6); iXY[1] = 58+(128*0); break;
		case 8: iXY[0] = 25+(64*7); iXY[1] = 58+(128*0); break;
		case 9: iXY[0] = 25+(64*8); iXY[1] = 58+(128*0); break;
		case 10: iXY[0] = 25+(64*9); iXY[1] = 58+(128*0); break;
		case 11: iXY[0] = 25+(64*0); iXY[1] = 58+(128*1); break;
		case 12: iXY[0] = 25+(64*1); iXY[1] = 58+(128*1); break;
		case 13: iXY[0] = 25+(64*2); iXY[1] = 58+(128*1); break;
		case 14: iXY[0] = 25+(64*3); iXY[1] = 58+(128*1); break;
		case 15: iXY[0] = 25+(64*4); iXY[1] = 58+(128*1); break;
		case 16: iXY[0] = 25+(64*5); iXY[1] = 58+(128*1); break;
		case 17: iXY[0] = 25+(64*6); iXY[1] = 58+(128*1); break;
		case 18: iXY[0] = 25+(64*7); iXY[1] = 58+(128*1); break;
		case 19: iXY[0] = 25+(64*8); iXY[1] = 58+(128*1); break;
		case 20: iXY[0] = 25+(64*9); iXY[1] = 58+(128*1); break;
		case 21: iXY[0] = 25+(64*0); iXY[1] = 58+(128*2); break;
		case 22: iXY[0] = 25+(64*1); iXY[1] = 58+(128*2); break;
		case 23: iXY[0] = 25+(64*2); iXY[1] = 58+(128*2); break;
		case 24: iXY[0] = 25+(64*3); iXY[1] = 58+(128*2); break;
		case 25: iXY[0] = 25+(64*4); iXY[1] = 58+(128*2); break;
		case 26: iXY[0] = 25+(64*5); iXY[1] = 58+(128*2); break;
		case 27: iXY[0] = 25+(64*6); iXY[1] = 58+(128*2); break;
		case 28: iXY[0] = 25+(64*7); iXY[1] = 58+(128*2); break;
		case 29: iXY[0] = 25+(64*8); iXY[1] = 58+(128*2); break;
		case 30: iXY[0] = 25+(64*9); iXY[1] = 58+(128*2); break;
		/*** event target ***/
		// TODO
		/*** custom tile ***/
		// TODO
		default:
			printf ("[ WARN ] Unknown tile location: %i!\n", iLocation);
			break;
	}
}
/*****************************************************************************/
int InArea (int iUpperLeftX, int iUpperLeftY,
	int iLowerRightX, int iLowerRightY)
/*****************************************************************************/
{
	if ((iUpperLeftX * iScale <= iXPos) &&
		(iLowerRightX * iScale >= iXPos) &&
		(iUpperLeftY * iScale <= iYPos) &&
		(iLowerRightY * iScale >= iYPos))
	{
		return (1);
	} else {
		return (0);
	}
}
/*****************************************************************************/
void PlaySound (char *sFile)
/*****************************************************************************/
{
	int iIndex;
	SDL_AudioSpec wave;
	Uint8 *data;
	Uint32 dlen;
	SDL_AudioCVT cvt;

	if (iNoAudio == 1) { return; }
	for (iIndex = 0; iIndex < NUM_SOUNDS; iIndex++)
	{
		if (sounds[iIndex].dpos == sounds[iIndex].dlen)
		{
			break;
		}
	}
	if (iIndex == NUM_SOUNDS) { return; }

	if (SDL_LoadWAV (sFile, &wave, &data, &dlen) == NULL)
	{
		printf ("[FAILED] Could not load %s: %s!\n", sFile, SDL_GetError());
		exit (EXIT_ERROR);
	}
	SDL_BuildAudioCVT (&cvt, wave.format, wave.channels, wave.freq, AUDIO_S16, 2,
		44100);
	/*** The "+ 1" is a workaround for SDL bug #2274. ***/
	cvt.buf = (Uint8 *)malloc (dlen * (cvt.len_mult + 1));
	memcpy (cvt.buf, data, dlen);
	cvt.len = dlen;
	SDL_ConvertAudio (&cvt);
	SDL_FreeWAV (data);

	if (sounds[iIndex].data)
	{
		free (sounds[iIndex].data);
	}
	SDL_LockAudio();
	sounds[iIndex].data = cvt.buf;
	sounds[iIndex].dlen = cvt.len_cvt;
	sounds[iIndex].dpos = 0;
	SDL_UnlockAudio();
}
/*****************************************************************************/
int BrokenRoomLinks (int iPrint)
/*****************************************************************************/
{
	int iBroken;
	int iTemp;

	for (iTemp = 1; iTemp <= ROOMS; iTemp++)
	{
		iDone[iTemp] = 0;
		arRoomLinksBroken[iTemp][1] = 0;
		arRoomLinksBroken[iTemp][2] = 0;
		arRoomLinksBroken[iTemp][3] = 0;
		arRoomLinksBroken[iTemp][4] = 0;
	}
	CheckSides (iKidRoom, 0, 0);
	iBroken = 0;

	for (iTemp = 1; iTemp <= ROOMS; iTemp++)
	{
		/*** If the room is in use... ***/
		if (iDone[iTemp] == 1)
		{
			/*** check left ***/
			if (arRoomLinks[iTemp][1] != 0)
			{
				if ((arRoomLinks[iTemp][1] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][1]][2] != iTemp))
				{
					arRoomLinksBroken[iTemp][1] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The left of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
			/*** check right ***/
			if (arRoomLinks[iTemp][2] != 0)
			{
				if ((arRoomLinks[iTemp][2] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][2]][1] != iTemp))
				{
					arRoomLinksBroken[iTemp][2] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The right of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
			/*** check up ***/
			if (arRoomLinks[iTemp][3] != 0)
			{
				if ((arRoomLinks[iTemp][3] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][3]][4] != iTemp))
				{
					arRoomLinksBroken[iTemp][3] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The top of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
			/*** check down ***/
			if (arRoomLinks[iTemp][4] != 0)
			{
				if ((arRoomLinks[iTemp][4] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][4]][3] != iTemp))
				{
					arRoomLinksBroken[iTemp][4] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The bottom of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
		}
	}

	return (iBroken);
}
/*****************************************************************************/
void CheckSides (int iRoom, int iX, int iY)
/*****************************************************************************/
{
	if (iX < iMinX) { iMinX = iX; }
	if (iY < iMinY) { iMinY = iY; }
	if (iX > iMaxX) { iMaxX = iX; }
	if (iY > iMaxY) { iMaxY = iY; }

	iDone[iRoom] = 1;

	if ((arRoomLinks[iRoom][1] != 0) &&
		(iDone[arRoomLinks[iRoom][1]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][1], iX - 1, iY); }

	if ((arRoomLinks[iRoom][2] != 0) &&
		(iDone[arRoomLinks[iRoom][2]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][2], iX + 1, iY); }

	if ((arRoomLinks[iRoom][3] != 0) &&
		(iDone[arRoomLinks[iRoom][3]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][3], iX, iY - 1); }

	if ((arRoomLinks[iRoom][4] != 0) &&
		(iDone[arRoomLinks[iRoom][4]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][4], iX, iY + 1); }
}
/*****************************************************************************/
void Prev (void)
/*****************************************************************************/
{
	int iToLoad;

	if (iCurLevel != 1)
	{
		iToLoad = iCurLevel - 1;
		if (iToLoad == -1) { iToLoad = MAX_LEVEL; }
		LoadLevel (iToLoad);
		ShowScreen (iScreen);
		PlaySound ("wav/level_change.wav");
	}
}
/*****************************************************************************/
void Next (void)
/*****************************************************************************/
{
	int iToLoad;

	if (iCurLevel != 0)
	{
		iToLoad = iCurLevel + 1;
		if (iToLoad == 16) { iToLoad = 0; }
		LoadLevel (iToLoad);
		ShowScreen (iScreen);
		PlaySound ("wav/level_change.wav");
	}
}
/*****************************************************************************/
void DisplayText (int iStartX, int iStartY, int iFontSize,
	char arText[9 + 2][MAX_TEXT + 2], int iLines, TTF_Font *font,
	SDL_Color back, SDL_Color fore)
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iLoopLine;

	for (iLoopLine = 0; iLoopLine < iLines; iLoopLine++)
	{
		if (strcmp (arText[iLoopLine], "") != 0)
		{
			message = TTF_RenderText_Shaded (font,
				arText[iLoopLine], fore, back);
			messaget = SDL_CreateTextureFromSurface (ascreen, message);
			if ((strcmp (arText[iLoopLine], "single tile (change or select)") == 0)
				|| (strcmp (arText[iLoopLine], "entire room (clear or fill)") == 0) ||
				(strcmp (arText[iLoopLine], "entire level (randomize or fill)") == 0))
			{
				offset.x = iStartX + 20;
			} else {
				offset.x = iStartX;
			}
			offset.y = iStartY + (iLoopLine * (iFontSize + 4));
			offset.w = message->w; offset.h = message->h;
			CustomRenderCopy (messaget, "message", NULL, ascreen, &offset);
			SDL_DestroyTexture (messaget); SDL_FreeSurface (message);
		}
	}
}
/*****************************************************************************/
void CustomRenderCopy (SDL_Texture* src, char *sImageInfo, SDL_Rect* srcrect,
	SDL_Renderer* dst, SDL_Rect *dstrect)
/*****************************************************************************/
{
	SDL_Rect stuff;

	stuff.x = dstrect->x * iScale;
	stuff.y = dstrect->y * iScale;
	if (srcrect != NULL) /*** image ***/
	{
		stuff.w = dstrect->w * iScale;
		stuff.h = dstrect->h * iScale;
	} else { /*** font ***/
		stuff.w = dstrect->w;
		stuff.h = dstrect->h;
	}
	if (SDL_RenderCopy (dst, src, srcrect, &stuff) != 0)
	{
		printf ("[ WARN ] SDL_RenderCopy (%s): %s!\n",
			sImageInfo, SDL_GetError());
	}
}
/*****************************************************************************/
void ChangeEvent (int iAmount, int iChangePos)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
int OnLevelBar (void)
/*****************************************************************************/
{
	if (InArea (28, 3, 28 + 575, 3 + 19) == 1)
		{ return (1); } else { return (0); }
}
/*****************************************************************************/
